import sqlite3
import json
from pathlib import Path
import logging
from datetime import datetime, timezone, timedelta
import os

log = logging.getLogger(__name__)

class CustomJsonEncoder(json.JSONEncoder):
    """
    Custom JSON encoder to handle datetime objects.
    """
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        return super().default(obj)

class Cache:
    def __init__(self, db_path: Path):
        self.db_path = db_path
        self.initialize_database()

    def _get_conn(self):
        """Get a new database connection."""
        try:
            conn = sqlite3.connect(self.db_path)
            conn.row_factory = sqlite3.Row
            return conn
        except sqlite3.Error as e:
            log.error(f"Database connection failed: {e}", exc_info=True)
            return None

    def initialize_database(self):
        """Create tables if they don't exist. Called once at startup."""
        conn = self._get_conn()
        if not conn:
            return
        try:
            cursor = conn.cursor()
            cursor.execute("""
                CREATE TABLE IF NOT EXISTS posts (
                    id TEXT PRIMARY KEY,
                    timeline_id TEXT NOT NULL,
                    created_at TEXT NOT NULL,
                    data TEXT NOT NULL
                )
            """)
            cursor.execute("CREATE INDEX IF NOT EXISTS idx_timeline_id ON posts (timeline_id, id)")
            conn.commit()
        except sqlite3.Error as e:
            log.error(f"Failed to create tables: {e}", exc_info=True)
        finally:
            if conn:
                conn.close()

    def bulk_insert_posts(self, timeline_id: str, posts: list):
        """Bulk insert posts into the database."""
        if not posts:
            return
        conn = self._get_conn()
        if not conn:
            return
        try:
            cursor = conn.cursor()
            posts_to_insert = []
            for post in posts:
                created_at_str = (post.get('reblog') or post).get('created_at')
                if isinstance(created_at_str, datetime):
                    created_at = created_at_str
                else:
                    created_at = datetime.fromisoformat(created_at_str.replace("Z", "+00:00"))
                
                posts_to_insert.append((
                    post['id'],
                    timeline_id,
                    created_at.isoformat(),
                    json.dumps(post, cls=CustomJsonEncoder)
                ))
            
            cursor.executemany(
                "INSERT OR REPLACE INTO posts (id, timeline_id, created_at, data) VALUES (?, ?, ?, ?)",
                posts_to_insert
            )
            conn.commit()
            log.info(f"Inserted/updated {len(posts_to_insert)} posts for timeline '{timeline_id}'")
        except sqlite3.Error as e:
            log.error(f"Failed to bulk insert posts: {e}", exc_info=True)
        finally:
            if conn:
                conn.close()

    def get_latest_post_timestamp(self, timeline_id: str) -> datetime | None:
        """Get the timestamp of the latest post in the cache for a timeline."""
        conn = self._get_conn()
        if not conn:
            return None
        try:
            cursor = conn.cursor()
            cursor.execute("SELECT MAX(created_at) FROM posts WHERE timeline_id = ?", (timeline_id,))
            row = cursor.fetchone()
            if row and row[0]:
                return datetime.fromisoformat(row[0])
            return None
        except sqlite3.Error as e:
            log.error(f"Failed to get latest post timestamp: {e}", exc_info=True)
            return None
        finally:
            if conn:
                conn.close()

    def get_posts(self, timeline_id: str, limit: int = 20, max_id: str = None):
        """Get posts from the database, ordered by ID."""
        conn = self._get_conn()
        if not conn:
            return []
        try:
            cursor = conn.cursor()
            query = "SELECT data FROM posts WHERE timeline_id = ?"
            params = [timeline_id]

            if max_id:
                query += " AND id < ?"
                params.append(max_id)

            query += " ORDER BY id DESC LIMIT ?"
            params.append(limit)

            cursor.execute(query, params)
            rows = cursor.fetchall()
            return [json.loads(row['data']) for row in rows]
        except sqlite3.Error as e:
            log.error(f"Failed to get posts: {e}", exc_info=True)
            return []
        finally:
            if conn:
                conn.close()

    def prune_image_cache(self, days: int = 30) -> int:
        """Prune the image cache of files older than a certain number of days."""
        count = 0
        cutoff = datetime.now(timezone.utc) - timedelta(days=days)
        
        for filename in os.listdir(config.image_cache_dir):
            file_path = config.image_cache_dir / filename
            try:
                if file_path.is_file():
                    modified_time = datetime.fromtimestamp(file_path.stat().st_mtime, tz=timezone.utc)
                    if modified_time < cutoff:
                        file_path.unlink()
                        count += 1
            except Exception as e:
                log.error(f"Error pruning cache file {file_path}: {e}", exc_info=True)
        
        log.info(f"Pruned {count} items from the image cache.")
        return count

# Setup a global cache instance
from mastui.config import config
config_dir = Path.home() / ".config" / "mastui"
config_dir.mkdir(parents=True, exist_ok=True)
cache_db_path = config_dir / "cache.db"
cache = Cache(cache_db_path)


from textual.screen import ModalScreen
from textual.widgets import Input, TabbedContent, TabPane, Static, LoadingIndicator
from textual.containers import Vertical
from textual import on, events
from textual.widget import Widget
from mastui.widgets import AccountResult, HashtagResult, StatusResult, SearchResult
from mastui.messages import ViewProfile
from mastui.thread import ThreadScreen
from mastui.hashtag_timeline import HashtagTimeline

class SearchScreen(ModalScreen):
    """A modal screen for searching."""

    BINDINGS = [
        ("escape", "app.pop_screen", "Close Search"),
        ("up", "cursor_up", "Cursor Up",),
        ("down", "cursor_down", "Cursor Down"),
        ("enter", "select_result", "Select"),
        ("p", "view_profile", "View Profile"),
    ]

    def __init__(self, api, **kwargs):
        super().__init__(**kwargs)
        self.api = api

    def compose(self):
        self.title = "Search"
        with Vertical(id="search-dialog"):
            yield Input(placeholder="Search for users, hashtags, or posts...", id="search-input")
            yield LoadingIndicator(classes="hidden")
            with TabbedContent(id="search-results"):
                with TabPane("Accounts", id="search-accounts"):
                    yield Static("Press Enter to search.", classes="search-status")
                with TabPane("Hashtags", id="search-hashtags"):
                    yield Static("Press Enter to search.", classes="search-status")
                with TabPane("Statuses", id="search-statuses"):
                    yield Static("Press Enter to search.", classes="search-status")

    def on_mount(self):
        """Focus the search input when the screen is mounted."""
        self.query_one("#search-input").focus()

    @on(Input.Submitted)
    def on_input_submitted(self, event: Input.Submitted) -> None:
        """Handle the search input being submitted."""
        query = event.value
        if not query:
            return

        # Add a hashtag if we're on the hashtag tab
        active_tab = self.query_one(TabbedContent).active
        if active_tab == "search-hashtags" and not query.startswith("#"):
            query = f"#{query}"
        
        self.query_one(LoadingIndicator).remove_class("hidden")
        self.run_worker(lambda: self.do_search(query), exclusive=True, thread=True)

    def do_search(self, query: str):
        """Worker method to perform the search."""
        try:
            results = self.api.search(query)
            self.app.call_from_thread(self.render_results, results)
        except Exception as e:
            self.app.notify(f"Error searching: {e}", severity="error")
            self.query_one(LoadingIndicator).add_class("hidden")

    def render_results(self, results: dict):
        """Render the search results."""
        self.query_one(LoadingIndicator).add_class("hidden")
        accounts_pane = self.query_one("#search-accounts", TabPane)
        hashtags_pane = self.query_one("#search-hashtags", TabPane)
        statuses_pane = self.query_one("#search-statuses", TabPane)

        accounts_pane.query("*").remove()
        hashtags_pane.query("*").remove()
        statuses_pane.query("*").remove()

        if results.get("accounts"):
            for account in results["accounts"]:
                accounts_pane.mount(AccountResult(account))
        else:
            accounts_pane.mount(Static("No account results.", classes="search-status"))

        if results.get("hashtags"):
            for hashtag in results["hashtags"]:
                hashtags_pane.mount(HashtagResult(hashtag))
        else:
            hashtags_pane.mount(Static("No hashtag results.", classes="search-status"))

        if results.get("statuses"):
            for status in results["statuses"]:
                statuses_pane.mount(StatusResult(status))
        else:
            statuses_pane.mount(Static("No status results.", classes="search-status"))

    @on(events.Click, ".search-result")
    def on_search_result_click(self, event: events.Click) -> None:
        """Handle a click on a search result."""
        self.select_result(event.widget)

    def action_cursor_up(self) -> None:
        """Move the cursor up."""
        self.query_one(TabbedContent).focus_previous(SearchResult)

    def action_cursor_down(self) -> None:
        """Move the cursor down."""
        self.query_one(TabbedContent).focus_next(SearchResult)

    def action_select_result(self) -> None:
        """Select the currently focused result."""
        focused = self.query_one("*:focus")
        if isinstance(focused, SearchResult):
            self.select_result(focused)

    def action_view_profile(self) -> None:
        """View the profile of the author of the focused post."""
        focused = self.query_one("*:focus")
        if isinstance(focused, AccountResult):
            self.select_result(focused)
        elif isinstance(focused, StatusResult):
            self.dismiss()
            self.post_message(ViewProfile(focused.status["account"]["id"]))

    def select_result(self, result_widget: Widget):
        """Handle a search result being selected."""
        if isinstance(result_widget, AccountResult):
            self.dismiss()
            self.post_message(ViewProfile(result_widget.account["id"]))
        elif isinstance(result_widget, StatusResult):
            self.dismiss()
            self.app.push_screen(ThreadScreen(result_widget.status["id"], self.app.api))
        elif isinstance(result_widget, HashtagResult):
            self.dismiss()
            self.app.push_screen(HashtagTimeline(result_widget.hashtag["name"], self.app.api))



from textual.screen import Screen
from textual.widgets import Static
from textual.containers import Vertical
from rich.panel import Panel
from importlib import metadata
import os
import toml
import logging

log = logging.getLogger(__name__)

class SplashScreen(Screen):
    """A splash screen with the app name, version, and logo."""

    def get_version(self):
        """
        Reads the version from the installed package metadata,
        falling back to pyproject.toml for development.
        """
        try:
            # For installed package
            return metadata.version("mastui")
        except metadata.PackageNotFoundError as e:
            log.error(f"Could not get version from metadata: {e}", exc_info=True)
            # For development environment
            pyproject_path = os.path.join(os.path.dirname(__file__), "..", "pyproject.toml")
            if os.path.exists(pyproject_path):
                with open(pyproject_path) as f:
                    pyproject_data = toml.load(f)
                return pyproject_data["tool"]["poetry"]["version"]
        return "unknown"

    def compose(self) -> None:
        logo = r"""
            [bold cyan]
            
 888b     d888                   888             d8b 
 8888b   d8888                   888             Y8P 
 88888b.d88888                   888                 
 888Y88888P888  8888b.  .d8888b  888888 888  888 888 
 888 Y888P 888     "88b 88K      888    888  888 8K8 
 888  Y8P  888 .d888888 "Y8888b. 888    888  888 8I8 
 888   "   888 888  888      X88 Y88b.  Y88b 888 8M8 
 888       888 "Y888888  88888P'  "Y888  "Y88888 888 
            [/bold cyan]
            """
        version = self.get_version()
        yield Vertical(
            Static(Panel(logo, border_style="dim"), id="logo"),
            Static(f"Mastui v{version}", id="version"),
            Static("Loading...", id="splash-status"),
            id="splash-container",
        )

    def on_mount(self) -> None:
        self.set_interval(0.3, self.update_loading_text)

    def update_loading_text(self) -> None:
        status = self.query_one("#splash-status")
        text = status.renderable
        if text.endswith("..."):
            status.update(text[:-3])
        else:
            status.update(text + ".")

    def update_status(self, message: str) -> None:
        """Update the status message on the splash screen."""
        try:
            self.query_one("#splash-status").update(message)
        except Exception as e:
            log.error(f"Could not update splash screen status: {e}", exc_info=True)




from textual.screen import ModalScreen
from textual.widgets import Static
from textual.containers import Vertical
from rich.panel import Panel
from rich.markdown import Markdown
from rich.text import Text
from mastui.utils import to_markdown
from mastui.image import ImageWidget
from mastui.config import config
import logging

log = logging.getLogger(__name__)

class ProfileScreen(ModalScreen):
    """A modal screen to display a user profile."""

    BINDINGS = [
        ("f", "follow", "Follow/Unfollow"),
        ("escape", "dismiss", "Dismiss")
    ]

    def __init__(self, account_id: str, api, **kwargs) -> None:
        super().__init__(**kwargs)
        self.account_id = account_id
        self.api = api
        self.profile = None

    def compose(self):
        with Vertical(id="profile-dialog"):
            yield Static("Loading profile...", classes="status-message")

    def on_mount(self):
        self.run_worker(self.load_profile, thread=True)

    def load_profile(self):
        """Load the user profile."""
        try:
            self.profile = self.api.account(self.account_id)
            # Check relationship
            relationships = self.api.account_relationships([self.account_id])
            if relationships:
                self.profile['following'] = relationships[0]['following']
            self.app.call_from_thread(self.render_profile)
        except Exception as e:
            log.error(f"Error loading profile: {e}", exc_info=True)
            self.app.notify(f"Error loading profile: {e}", severity="error")
            self.dismiss()

    def render_profile(self):
        """Render the profile."""
        profile = self.profile
        if not profile:
            return

        container = self.query_one("#profile-dialog")
        container.query("*").remove()

        header = f"[bold]{profile['display_name']}[/bold] (@{profile['acct']})"
        
        # Add following status to header
        if profile.get('following'):
            header += " [green](Following)[/green]"

        note_html = profile.get('note', '')
        note = Markdown(to_markdown(note_html)) if note_html else "No bio."
        
        stats = f"Following: {profile['following_count']} | Followers: {profile['followers_count']} | Posts: {profile['statuses_count']}"

        fields_text = ""
        if profile.get('fields'):
            for field in profile['fields']:
                fields_text += f"**{field['name']}:** {to_markdown(field['value'])}\n"

        if config.image_support:
            container.mount(ImageWidget(profile['avatar'], config.image_renderer, id="profile-avatar"))

        container.mount(
            Static(Panel(header, border_style="dim")),
            Static(Panel(note, title="Bio"), id="profile-bio"),
            Static(Panel(Markdown(fields_text), title="Links"), id="profile-links"),
            Static(stats, id="profile-stats")
        )

    def action_follow(self):
        """Follow or unfollow the user."""
        if not self.profile:
            return

        try:
            if self.profile.get('following'):
                self.api.account_unfollow(self.account_id)
                self.app.notify(f"Unfollowed @{self.profile['acct']}")
            else:
                self.api.account_follow(self.account_id)
                self.app.notify(f"Followed @{self.profile['acct']}")
            
            # Reload profile to update status
            self.run_worker(self.load_profile, thread=True)

        except Exception as e:
            log.error(f"Error following/unfollowing: {e}", exc_info=True)
            self.app.notify(f"Error: {e}", severity="error")



from textual.app import App, ComposeResult
from textual.widgets import Header, Footer
from textual import on, events
from textual.screen import ModalScreen
from mastui.login import LoginScreen
from mastui.post import PostScreen
from mastui.reply import ReplyScreen
from mastui.splash import SplashScreen
from mastui.mastodon_api import get_api
from mastui.timeline import Timelines, Timeline
from mastui.widgets import Post, LikePost, BoostPost, VoteOnPoll
from mastui.thread import ThreadScreen
from mastui.profile import ProfileScreen
from mastui.config_screen import ConfigScreen
from mastui.help_screen import HelpScreen
from mastui.search_screen import SearchScreen
from mastui.logging_config import setup_logging
from mastui.retro import retro_theme
import logging
import argparse
import os
from mastui.config import config
from mastui.messages import (
    PostStatusUpdate,
    ActionFailed,
    TimelineData,
    FocusNextTimeline,
    FocusPreviousTimeline,
    ViewProfile,
    ResumeTimers,
)
from mastui.cache import cache
from mastodon.errors import MastodonAPIError

# Set up logging
log = logging.getLogger(__name__)


# Get the absolute path to the CSS file
css_path = os.path.join(os.path.dirname(__file__), "app.css")


class Mastui(App):
    """A Textual app to interact with Mastodon."""

    BINDINGS = [
        ("d", "toggle_dark", "Toggle dark mode"),
        ("r", "refresh_timelines", "Refresh timelines"),
        ("c", "compose_post", "Compose post"),
        ("p", "view_profile", "View profile"),
        ("a", "reply_to_post", "Reply to post"),
        ("o", "open_options", "Options"),
        ("/", "search", "Search"),
        ("?", "show_help", "Help"),
        ("q", "quit", "Quit"),
        ("l", "like_post", "Like post"),
        ("b", "boost_post", "Boost post"),
        ("up", "scroll_up", "Scroll up"),
        ("down", "scroll_down", "Scroll down"),
    ]
    CSS_PATH = css_path
    initial_data = None
    max_characters = 500 # Default value
    log_file_path: str | None = None

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header()
        yield Footer()

    def on_mount(self) -> None:
        """Called when the app is mounted."""
        self.register_theme(retro_theme)
        self.theme = config.theme
        self.theme_changed_signal.subscribe(self, self.on_theme_changed)
        self.push_screen(SplashScreen())
        self.api = get_api()
        if self.api:
            self.run_worker(self.fetch_instance_info, thread=True, exclusive=True)
            if config.auto_prune_cache:
                self.run_worker(self.prune_cache, thread=True, exclusive=True)
            self.set_timer(2, self.show_timelines)
        else:
            self.call_later(self.show_login_screen)

    def fetch_instance_info(self):
        """Fetches instance information from the API."""
        try:
            instance = self.api.instance()
            self.max_characters = instance['configuration']['statuses']['max_characters']
        except Exception as e:
            log.error(f"Error fetching instance info: {e}", exc_info=True)
            self.notify("Could not fetch instance information.", severity="error")

    def prune_cache(self):
        """Prunes the image cache."""
        try:
            count = cache.prune_image_cache()
            if count > 0:
                self.notify(f"Pruned {count} items from the image cache.")
        except Exception as e:
            log.error(f"Error pruning cache: {e}", exc_info=True)
            self.notify("Error pruning image cache.", severity="error")

    def on_theme_changed(self, event) -> None:
        """Called when the app's theme is changed."""
        new_theme = event.name
        config.theme = new_theme
        if "light" in new_theme:
            config.preferred_light_theme = new_theme
        else:
            config.preferred_dark_theme = new_theme
        config.save_config()

    def show_login_screen(self):
        if isinstance(self.screen, SplashScreen):
            self.pop_screen()
        self.push_screen(LoginScreen(), self.on_login)

    def on_login(self, api) -> None:
        """Called when the login screen is dismissed."""
        log.info("Login successful.")
        self.api = api
        self.run_worker(self.fetch_instance_info, thread=True, exclusive=True)
        self.push_screen(SplashScreen())
        self.set_timer(2, self.show_timelines)

    def show_timelines(self):
        if isinstance(self.screen, SplashScreen):
            self.pop_screen()
        log.info("Showing timelines...")
        self.mount(Timelines())
        self.call_later(self.check_layout_mode)

    @on(events.Resize)
    def on_resize(self, event: events.Resize) -> None:
        """Called when the app is resized."""
        self.check_layout_mode()

    def check_layout_mode(self) -> None:
        """Check and apply the layout mode based on screen size."""
        is_narrow = config.force_single_column or self.size.width < self.size.height
        try:
            timelines = self.query_one(Timelines)
            timelines.set_class(is_narrow, "single-column-mode")
        except Exception as e:
            log.debug(f"Could not apply layout mode, Timelines widget not found: {e}")

    

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        if "light" in self.theme:
            self.theme = config.preferred_dark_theme
        else:
            self.theme = config.preferred_light_theme

    def action_open_options(self) -> None:
        """An action to open the options screen."""
        if isinstance(self.screen, ModalScreen):
            return
        self.pause_timers()
        self.push_screen(ConfigScreen(), self.on_config_screen_dismiss)

    def action_show_help(self) -> None:
        """An action to show the help screen."""
        if isinstance(self.screen, ModalScreen):
            return
        self.pause_timers()
        self.push_screen(HelpScreen(), self.on_help_screen_dismiss)

    def action_search(self) -> None:
        """An action to open the search screen."""
        if isinstance(self.screen, ModalScreen):
            return
        self.pause_timers()
        self.push_screen(SearchScreen(api=self.api), self.on_search_screen_dismiss)

    def on_search_screen_dismiss(self, _) -> None:
        """Called when the search screen is dismissed."""
        self.resume_timers()

    def on_help_screen_dismiss(self) -> None:
        """Called when the help screen is dismissed."""
        self.resume_timers()

    def on_config_screen_dismiss(self, result: bool) -> None:
        """Called when the config screen is dismissed."""
        self.resume_timers()
        if result:
            self.query_one(Timelines).remove()
            self.mount(Timelines())
            self.call_later(self.check_layout_mode)

    def action_refresh_timelines(self) -> None:
        """An action to refresh the timelines."""
        log.info("Refreshing all timelines...")
        for timeline in self.query(Timeline):
            timeline.refresh_posts()

    def action_compose_post(self) -> None:
        """An action to compose a new post."""
        if isinstance(self.screen, ModalScreen):
            return
        self.pause_timers()
        self.push_screen(PostScreen(max_characters=self.max_characters), self.on_post_screen_dismiss)

    def action_reply_to_post(self) -> None:
        focused = self.query("Timeline:focus")
        if focused:
            self.pause_timers()
            focused.first().reply_to_post()

    def on_post_screen_dismiss(self, result: dict) -> None:
        """Called when the post screen is dismissed."""
        self.resume_timers()
        if result:
            try:
                log.info("Sending post...")
                self.api.status_post(
                    status=result["content"],
                    spoiler_text=result["spoiler_text"],
                    language=result["language"],
                    poll=result["poll"],
                )
                log.info("Post sent successfully.")
                self.notify("Post sent successfully!", severity="information")
                self.action_refresh_timelines()
            except Exception as e:
                log.error(f"Error sending post: {e}", exc_info=True)
                self.notify(f"Error sending post: {e}", severity="error")

    def on_reply_screen_dismiss(self, result: dict) -> None:
        """Called when the reply screen is dismissed."""
        self.resume_timers()
        if result:
            try:
                log.info(f"Sending reply to post {result['in_reply_to_id']}...")
                self.api.status_post(
                    status=result["content"],
                    spoiler_text=result["spoiler_text"],
                    language=result["language"],
                    in_reply_to_id=result["in_reply_to_id"],
                )
                log.info("Reply sent successfully.")
                self.notify("Reply sent successfully!", severity="information")
                self.action_refresh_timelines()
            except Exception as e:
                log.error(f"Error sending reply: {e}", exc_info=True)
                self.notify(f"Error sending reply: {e}", severity="error")

    @on(LikePost)
    def handle_like_post(self, message: LikePost):
        self.run_worker(lambda: self.do_like_post(message.post_id, message.favourited), exclusive=True, thread=True)

    def do_like_post(self, post_id: str, favourited: bool):
        try:
            if favourited:
                post_data = self.api.status_unfavourite(post_id)
            else:
                post_data = self.api.status_favourite(post_id)
            self.post_message(PostStatusUpdate(post_data))
        except Exception as e:
            log.error(f"Error liking/unliking post {post_id}: {e}", exc_info=True)
            self.post_message(ActionFailed(post_id))

    @on(BoostPost)
    def handle_boost_post(self, message: BoostPost):
        self.run_worker(lambda: self.do_boost_post(message.post_id), exclusive=True, thread=True)

    def do_boost_post(self, post_id: str):
        try:
            post_data = self.api.status_reblog(post_id)
            self.post_message(PostStatusUpdate(post_data))
        except Exception as e:
            log.error(f"Error boosting post {post_id}: {e}", exc_info=True)
            self.post_message(ActionFailed(post_id))

    @on(VoteOnPoll)
    def handle_vote_on_poll(self, message: VoteOnPoll):
        self.run_worker(lambda: self.do_vote_on_poll(message.poll_id, message.choice, message.timeline_id, message.post_id), exclusive=True, thread=True)

    def do_vote_on_poll(self, poll_id: str, choice: int, timeline_id: str, post_id: str):
        try:
            # The API returns the updated post object after voting
            updated_post_data = self.api.poll_vote(poll_id, [choice])
            
            # Update the cache with the new post data
            cache.bulk_insert_posts(timeline_id, [updated_post_data])
            
            self.post_message(PostStatusUpdate(updated_post_data))
            self.notify("Vote cast successfully!", severity="information")
        except MastodonAPIError as e:
            if "You have already voted on this poll" in str(e):
                log.info(f"User already voted on poll {poll_id}. Fetching latest post state for post {post_id}.")
                try:
                    # Fetch the latest post data to get the correct poll state
                    updated_post_data = self.api.status(post_id)
                    cache.bulk_insert_posts(timeline_id, [updated_post_data])
                    self.post_message(PostStatusUpdate(updated_post_data))
                except Exception as fetch_e:
                    log.error(f"Error fetching post {post_id} after 'already voted' error: {fetch_e}", exc_info=True)
                    self.notify("Could not refresh poll state.", severity="error")
            else:
                log.error(f"Error voting on poll {poll_id}: {e}", exc_info=True)
                self.notify(f"Error casting vote: {e}", severity="error")
                self.action_refresh_timelines() # Fallback
        except Exception as e:
            log.error(f"Unexpected error voting on poll {poll_id}: {e}", exc_info=True)
            self.notify(f"Error casting vote: {e}", severity="error")
            self.action_refresh_timelines()

    def on_post_status_update(self, message: PostStatusUpdate) -> None:
        updated_post_data = message.post_data
        target_post = updated_post_data.get("reblog") or updated_post_data
        target_id = target_post["id"]

        for container in [self.screen, *self.query(Timelines)]:
            for post_widget in container.query(Post):
                original_status = post_widget.post.get("reblog") or post_widget.post
                if original_status["id"] == target_id:
                    post_widget.update_from_post(updated_post_data)

    def on_action_failed(self, message: ActionFailed) -> None:
        for container in [self.screen, *self.query(Timelines)]:
            for post_widget in container.query(Post):
                original_status = post_widget.post.get("reblog") or post_widget.post
                if original_status["id"] == message.post_id:
                    post_widget.hide_spinner()

    @on(FocusNextTimeline)
    def on_focus_next_timeline(self, message: FocusNextTimeline) -> None:
        timelines = self.query(Timeline)
        for i, timeline in enumerate(timelines):
            if timeline.has_focus:
                timelines[(i + 1) % len(timelines)].focus()
                return

    @on(FocusPreviousTimeline)
    def on_focus_previous_timeline(self, message: FocusPreviousTimeline) -> None:
        timelines = self.query(Timeline)
        for i, timeline in enumerate(timelines):
            if timeline.has_focus:
                timelines[(i - 1) % len(timelines)].focus()
                return

    @on(ViewProfile)
    def on_view_profile(self, message: ViewProfile) -> None:
        if isinstance(self.screen, ModalScreen):
            return
        self.pause_timers()
        self.push_screen(ProfileScreen(message.account_id, self.api), self.on_profile_screen_dismiss)

    def on_profile_screen_dismiss(self, _) -> None:
        """Called when the profile screen is dismissed."""
        self.resume_timers()

    def pause_timers(self):
        """Pauses all timeline timers."""
        for timeline in self.query(Timeline):
            timeline.pause_timers()

    def resume_timers(self):
        """Resumes all timeline timers."""
        for timeline in self.query(Timeline):
            timeline.resume_timers()

    def action_like_post(self) -> None:
        focused = self.query("Timeline:focus")
        if focused:
            focused.first().like_post()

    def action_boost_post(self) -> None:
        focused = self.query("Timeline:focus")
        if focused:
            focused.first().boost_post()

    def action_scroll_up(self) -> None:
        focused = self.query("Timeline:focus")
        if focused:
            focused.first().scroll_up()

    def action_scroll_down(self) -> None:
        focused = self.query("Timeline:focus")
        if focused:
            focused.first().scroll_down()


def main():
    parser = argparse.ArgumentParser(description="A Textual app to interact with Mastodon.")
    parser.add_argument("--no-ssl-verify", action="store_false", dest="ssl_verify", help="Disable SSL verification.")
    parser.add_argument("--debug", action="store_true", help="Enable debug logging.")
    args = parser.parse_args()
    
    log_file_path = setup_logging(debug=args.debug)
    
    config.ssl_verify = args.ssl_verify
    app = Mastui()
    app.log_file_path = log_file_path
    
    try:
        app.run()
    finally:
        if app.log_file_path:
            print(f"Log file written to: {app.log_file_path}")


if __name__ == "__main__":
    main()



from textual.theme import Theme

retro_theme = Theme(
    name="retro-green",
    primary="#00ff00",
    secondary="#00cc00",
    accent="#00dd00",
    foreground="#00ff00",
    background="#000000",
    surface="#002200",
    panel="#003300",
    success="#55ff55",
    warning="#ffff00",
    error="#ff5555",
    dark=True,
)



from textual.screen import ModalScreen
from textual.widgets import (
    Button,
    Label,
    Input,
    Switch,
    Static,
    Select,
    Collapsible,
    Header,
)
from textual.containers import Grid, Vertical, Horizontal
from textual import on
from mastui.config import config
from mastui.cache import cache


class ConfigScreen(ModalScreen):
    """A modal screen for changing settings."""

    def compose(self):
        self.title = "Mastui Options"
        with Vertical(id="config-dialog"):
            yield Header(show_clock=False)

            with Collapsible(title="Timeline Visibility"):
                with Grid(classes="config-group-body"):
                    yield Label("Enable Home timeline?", classes="config-label")
                    yield Switch(
                        value=config.home_timeline_enabled, id="home_timeline_enabled"
                    )
                    yield Static()  # Spacer

                    yield Label(
                        "Enable Notifications timeline?", classes="config-label"
                    )
                    yield Switch(
                        value=config.notifications_timeline_enabled,
                        id="notifications_timeline_enabled",
                    )
                    yield Static()  # Spacer

                    yield Label("Enable Federated timeline?", classes="config-label")
                    yield Switch(
                        value=config.federated_timeline_enabled,
                        id="federated_timeline_enabled",
                    )
                    yield Static()  # Spacer

                    yield Label("Force single-column mode?", classes="config-label")
                    yield Switch(
                        value=config.force_single_column, id="force_single_column"
                    )
                    yield Static() # Spacer

            with Collapsible(title="Auto-Refresh (in minutes)"):
                with Grid(classes="config-group-body"):
                    yield Label("Auto-refresh home?", classes="config-label")
                    yield Switch(value=config.home_auto_refresh, id="home_auto_refresh")
                    yield Input(
                        str(config.home_auto_refresh_interval),
                        id="home_auto_refresh_interval",
                    )

                    yield Label("Auto-refresh notifications?", classes="config-label")
                    yield Switch(
                        value=config.notifications_auto_refresh,
                        id="notifications_auto_refresh",
                    )
                    yield Input(
                        str(config.notifications_auto_refresh_interval),
                        id="notifications_auto_refresh_interval",
                    )

                    yield Label("Auto-refresh federated?", classes="config-label")
                    yield Switch(
                        value=config.federated_auto_refresh, id="federated_auto_refresh"
                    )
                    yield Input(
                        str(config.federated_auto_refresh_interval),
                        id="federated_auto_refresh_interval",
                    )

            with Collapsible(title="Images & Cache"):
                with Grid(classes="config-group-body"):
                    yield Label("Show images?", classes="config-label")
                    yield Switch(value=config.image_support, id="image_support")
                    yield Select(
                        [
                            ("Auto", "auto"),
                            ("ANSI", "ansi"),
                            ("Sixel", "sixel"),
                            ("TGP (iTerm2)", "tgp"),
                        ],
                        value=config.image_renderer,
                        id="image_renderer",
                    )

                    yield Label(
                        "Auto-prune cache (older than 30 days)?", classes="config-label"
                    )
                    yield Switch(value=config.auto_prune_cache, id="auto_prune_cache")
                    yield Static()  # Spacer

            with Horizontal(id="config-buttons"):
                yield Button("Save", variant="primary", id="save")
                yield Button("Cancel", id="cancel")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "save":
            self.save_settings()
            self.dismiss(True)
        else:
            self.dismiss(False)

    @on(Collapsible.Toggled)
    def on_collapsible_toggled(self, event: Collapsible.Toggled) -> None:
        """When a collapsible is opened, close the others."""
        if not event.collapsible.collapsed:
            for collapsible in self.query(Collapsible):
                if collapsible is not event.collapsible:
                    collapsible.collapsed = True

    def on_switch_changed(self, event: Switch.Changed) -> None:
        if event.switch.id == "auto_prune_cache" and event.value:
            self.app.prune_cache()

    def save_settings(self):
        """Saves the current settings to the config object."""
        config.home_auto_refresh = self.query_one("#home_auto_refresh").value
        config.home_auto_refresh_interval = int(
            self.query_one("#home_auto_refresh_interval").value
        )
        config.notifications_auto_refresh = self.query_one(
            "#notifications_auto_refresh"
        ).value
        config.notifications_auto_refresh_interval = int(
            self.query_one("#notifications_auto_refresh_interval").value
        )
        config.federated_auto_refresh = self.query_one("#federated_auto_refresh").value
        config.federated_auto_refresh_interval = int(
            self.query_one("#federated_auto_refresh_interval").value
        )
        config.image_support = self.query_one("#image_support").value
        config.image_renderer = self.query_one("#image_renderer").value
        config.auto_prune_cache = self.query_one("#auto_prune_cache").value
        config.home_timeline_enabled = self.query_one("#home_timeline_enabled").value
        config.notifications_timeline_enabled = self.query_one(
            "#notifications_timeline_enabled"
        ).value
        config.federated_timeline_enabled = self.query_one(
            "#federated_timeline_enabled"
        ).value
        config.force_single_column = self.query_one("#force_single_column").value
        config.save_config()



from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Button, Label, Input, Static, TextArea, Select, Header, Switch
from textual.containers import Grid, Horizontal, Vertical, VerticalScroll
from textual import on
from textual.css.query import NoMatches
from mastui.utils import LANGUAGE_OPTIONS
from mastui.widgets import PollChoice, RemovePollChoice, PollChoiceMounted
import logging

log = logging.getLogger(__name__)
from textual import on
from mastui.utils import LANGUAGE_OPTIONS

class PostScreen(ModalScreen):
    """A modal screen for composing a new post."""

    BINDINGS = [
        ("escape", "app.pop_screen", "Cancel Post"),
    ]

    def __init__(self, max_characters: int = 500, **kwargs):
        super().__init__(**kwargs)
        self.max_characters = max_characters

    def compose(self):
        self.title = "New Post"
        with Vertical(id="post_dialog"):
            yield Header(show_clock=False)
            with VerticalScroll(id="post_content_container"):
                yield TextArea(id="post_content", language="markdown")
                with Horizontal(id="post_options"):
                    yield Label("CW:", classes="post_option_label")
                    yield Input(placeholder="Content warning", id="cw_input")
                with Horizontal(id="post_language_container"):
                    yield Label("Language:", classes="post_option_label")
                    yield Select(LANGUAGE_OPTIONS, value="en", id="language_select")
                
                with Horizontal(id="add_poll_container"):
                    yield Label("Add Poll:", classes="post_option_label")
                    yield Switch(id="add_poll_switch")

                with Vertical(id="poll_creation_area", classes="hidden"):
                    yield Label("Poll Choices:")
                    yield Vertical(id="poll_choices_container")
                    yield Button("Add Choice", id="add_choice_button")
                    
                    yield Label("Poll Duration:")
                    yield Select(
                        [
                            ("5 minutes", 300),
                            ("30 minutes", 1800),
                            ("1 hour", 3600),
                            ("6 hours", 21600),
                            ("1 day", 86400),
                            ("3 days", 259200),
                            ("7 days", 604800),
                        ],
                        value=86400,
                        id="poll_duration_select",
                    )

            with Horizontal(id="post_buttons"):
                yield Label(f"{self.max_characters}", id="character_limit")
                yield Button("Post", variant="primary", id="post")
                yield Button("Cancel", id="cancel")

    def on_mount(self):
        self.query_one("#post_content").focus()
        self.update_character_limit()

    @on(Switch.Changed, "#add_poll_switch")
    def on_add_poll_switch_changed(self, event: Switch.Changed) -> None:
        """Toggle the poll creation area."""
        poll_area = self.query_one("#poll_creation_area")
        choices_container = self.query_one("#poll_choices_container")
        if event.value:
            poll_area.remove_class("hidden")
            # Add initial choices if none exist
            if not self.query(PollChoice):
                choices_container.mount(PollChoice())
                choices_container.mount(PollChoice())
                self.call_after_refresh(self.update_remove_buttons)
        else:
            poll_area.add_class("hidden")
            for choice in self.query(PollChoice):
                choice.remove()

    @on(Button.Pressed, "#add_choice_button")
    def on_add_choice_button_pressed(self, event: Button.Pressed) -> None:
        """Handle the 'Add Choice' button press."""
        self.query_one("#poll_choices_container").mount(PollChoice())
        self.call_after_refresh(self.update_remove_buttons)

    @on(RemovePollChoice)
    def on_remove_poll_choice(self, message: RemovePollChoice) -> None:
        """Handle the removal of a poll choice."""
        message.poll_choice_widget.remove()
        self.update_remove_buttons()

    @on(PollChoiceMounted)
    def on_poll_choice_mounted(self, message: PollChoiceMounted) -> None:
        """Handle a poll choice being mounted."""
        self.update_remove_buttons()

    def update_remove_buttons(self) -> None:
        """Enable or disable remove buttons based on the number of choices."""
        choices = self.query(PollChoice)
        can_remove = len(choices) > 2
        for choice in choices:
            try:
                choice.query_one(".remove-choice").disabled = not can_remove
            except NoMatches:
                pass

    @on(Input.Changed)
    @on(TextArea.Changed)
    def update_character_limit(self):
        """Updates the character limit."""
        content_len = len(self.query_one("#post_content").text)
        cw_len = len(self.query_one("#cw_input").value)
        remaining = self.max_characters - content_len - cw_len
        
        limit_label = self.query_one("#character_limit")
        limit_label.update(f"{remaining}")
        limit_label.set_class(remaining < 0, "character-limit-error")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "post":
            content = self.query_one("#post_content").text
            spoiler_text = self.query_one("#cw_input").value
            language = self.query_one("#language_select").value
            
            result = {
                "content": content,
                "spoiler_text": spoiler_text,
                "language": language,
                "poll": None,
            }

            if self.query_one("#add_poll_switch").value:
                choices = [
                    choice.query_one(Input).value
                    for choice in self.query(PollChoice)
                ]
                if any(choices): # Only add poll if there are choices
                    result["poll"] = {
                        "options": choices,
                        "expires_in": self.query_one("#poll_duration_select").value,
                    }

            self.dismiss(result)
        elif event.button.id == "cancel":
            self.dismiss(None)



from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Button, Static, TextArea, Input, Switch, Select, Label, Header
from textual.containers import Vertical, Horizontal, Grid, VerticalScroll
from rich.panel import Panel
from rich.markdown import Markdown
from rich import box
from textual import on
from mastui.utils import get_full_content_md, LANGUAGE_OPTIONS

class ReplyScreen(ModalScreen):
    """A modal screen for replying to a post."""

    BINDINGS = [
        ("escape", "app.pop_screen", "Cancel Reply"),
    ]

    def __init__(self, post_to_reply_to, max_characters: int = 500, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.post_to_reply_to = post_to_reply_to
        self.max_characters = max_characters

    def get_mentions(self):
        """Get mentions from the post being replied to."""
        mentions = {f"@{self.post_to_reply_to['account']['acct']}"}
        for mention in self.post_to_reply_to.get('mentions', []):
            mentions.add(f"@{mention['acct']}")
        return " ".join(sorted(list(mentions)))

    def compose(self) -> ComposeResult:
        self.title = "Reply to Post"
        with Vertical(id="reply_dialog"):
            yield Header(show_clock=False)
            with VerticalScroll(id="reply_content_container"):
                yield Static(
                    Panel(
                        Markdown(get_full_content_md(self.post_to_reply_to)),
                        title=f"Replying to @{self.post_to_reply_to['account']['acct']}",
                        box=box.ROUNDED,
                        padding=(0, 1),
                    ),
                    id="original_post_preview"
                )
                reply_text_area = TextArea(id="reply_content", language="markdown")
                reply_text_area.text = self.get_mentions() + " "
                yield reply_text_area
                with Horizontal(id="reply_options"):
                    yield Static("Content Warning:", classes="reply_option_label")
                    yield Switch(id="cw_switch")
                    yield Input(id="cw_input", placeholder="Spoiler text...", disabled=True)
                with Horizontal(id="reply_language_container"):
                    yield Static("Language:", classes="reply_option_label")
                    yield Select(LANGUAGE_OPTIONS, id="language_select", value="en")
            with Horizontal(id="reply_buttons"):
                yield Label(f"{self.max_characters}", id="character_limit")
                yield Button("Post Reply", variant="primary", id="post_button")
                yield Button("Cancel", id="cancel_button")

    def on_mount(self) -> None:
        """Set initial focus."""
        self.query_one("#reply_content").focus()
        self.query_one("#reply_content").cursor_location = (0, len(self.query_one("#reply_content").text))
        self.update_character_limit()

    @on(Input.Changed)
    @on(TextArea.Changed)
    def update_character_limit(self):
        """Updates the character limit."""
        content_len = len(self.query_one("#reply_content").text)
        cw_len = len(self.query_one("#cw_input").value)
        remaining = self.max_characters - content_len - cw_len
        
        limit_label = self.query_one("#character_limit")
        limit_label.update(f"{remaining}")
        limit_label.set_class(remaining < 0, "character-limit-error")

    def on_switch_changed(self, event: Switch.Changed) -> None:
        """Toggle the content warning input."""
        cw_input = self.query_one("#cw_input")
        if event.value:
            cw_input.disabled = False
            cw_input.focus()
        else:
            cw_input.disabled = True
            cw_input.value = ""

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button presses."""
        if event.button.id == "post_button":
            content = self.query_one("#reply_content").text
            cw_text = self.query_one("#cw_input").value
            language = self.query_one("#language_select").value
            
            if content:
                result = {
                    "content": content,
                    "spoiler_text": cw_text if self.query_one("#cw_switch").value else None,
                    "language": language,
                    "in_reply_to_id": self.post_to_reply_to['id']
                }
                self.dismiss(result)
            else:
                self.app.notify("Reply content cannot be empty.", severity="error")

        elif event.button.id == "cancel_button":
            self.dismiss(None)



import html2text
import re
from datetime import datetime, timezone
from dateutil.parser import parse
from bs4 import BeautifulSoup

LANGUAGE_OPTIONS = [
    ("Chinese", "zh"),
    ("Danish", "da"),
    ("English", "en"),
    ("French", "fr"),
    ("German", "de"),
    ("Japanese", "ja"),
    ("Korean", "ko"),
    ("Spanish", "es"),
]

MARKDOWN_LINK_REGEX = re.compile(r'\\[^\\]+\\]\(([^)]+)\\\)')

def markdown_links_to_html(text: str) -> str:
    """Converts Markdown-style links in a string to HTML <a> tags."""
    if not text:
        return ""
    return MARKDOWN_LINK_REGEX.sub(lambda m: f'<a href="{m.group(2)}">{m.group(1)}</a>', text)

def to_markdown(html: str) -> str:
    """
    Converts mixed HTML/Markdown content to terminal-friendly markdown.
    """
    if not html:
        return ""

    # First, parse with BeautifulSoup to handle HTML and decode entities.
    soup = BeautifulSoup(html, 'html.parser')
    
    # Convert the parsed soup back to a string.
    cleaned_html = str(soup)

    # Now, convert any remaining Markdown-style links to HTML <a> tags.
    html_with_links = markdown_links_to_html(cleaned_html)

    # Finally, convert the fully-HTML content to terminal-friendly text.
    h = html2text.HTML2Text()
    h.ignore_links = False
    h.body_width = 0 # Don't wrap lines
    return h.handle(html_with_links)

def get_full_content_md(status: dict) -> str:
    """Gets the full markdown content for a status, including media."""
    if not status:
        return ""

    html_content = status.get("content") or status.get("note") or ""
    
    content_md = to_markdown(html_content)

    if status.get("media_attachments"):
        media_infos = []
        for media in status["media_attachments"]:
            media_type = media.get("type", "media").capitalize()
            description = media.get("description")

            if description:
                media_infos.append(f"[{media_type} showing: {description}]")
            else:
                media_infos.append(f"[{media_type} attached]")

        if content_md.strip():
            content_md += "\n\n" + "\n".join(media_infos)
        else:
            content_md = "\n".join(media_infos)

    return content_md



def format_datetime(dt_obj):
    """Formats a datetime string or object into YYYY-MM-DD HH:MM."""
    if isinstance(dt_obj, str):
        dt = parse(dt_obj)
    else:
        dt = dt_obj

    # Convert to local timezone for display
    local_dt = dt.astimezone()

    return local_dt.strftime("%Y-%m-%d %H:%M")



import html2text
from textual.widgets import Static, LoadingIndicator, RadioSet, RadioButton, Button, Input
from textual.widget import Widget
from rich.panel import Panel
from rich.markdown import Markdown
from rich import box
from textual.containers import VerticalScroll, Vertical, Horizontal
from textual import events, on
from textual.message import Message
from mastui.utils import get_full_content_md, format_datetime, to_markdown
from mastui.reply import ReplyScreen
from mastui.image import ImageWidget
from mastui.config import config
from mastui.messages import SelectPost, VoteOnPoll
import logging
from datetime import datetime

log = logging.getLogger(__name__)


class PostMessage(Message):
    """A message relating to a post."""

    def __init__(self, post_id: str) -> None:
        self.post_id = post_id
        super().__init__()


class LikePost(PostMessage):
    """A message to like a post."""

    def __init__(self, post_id: str, favourited: bool):
        super().__init__(post_id)
        self.favourited = favourited


class BoostPost(PostMessage):
    """A message to boost a post."""

    pass


class RemovePollChoice(Message):
    """A message to remove a poll choice."""
    def __init__(self, poll_choice_widget: Widget) -> None:
        self.poll_choice_widget = poll_choice_widget
        super().__init__()


class PollChoiceMounted(Message):
    """A message to indicate that a poll choice has been mounted."""
    pass


class PollChoice(Horizontal):
    """A widget for a single poll choice."""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_class("poll-choice")

    def compose(self):
        yield Input(placeholder="Choice...")
        yield Button("X", variant="error", classes="remove-choice", disabled=True)

    def on_mount(self) -> None:
        """When the widget is mounted, tell the parent to check button states."""
        self.post_message(PollChoiceMounted())

    @on(Button.Pressed, ".remove-choice")
    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.post_message(RemovePollChoice(self))


class PollWidget(Vertical):
    """A widget to display a poll."""

    def __init__(self, poll: dict, timeline_id: str, post_id: str, **kwargs):
        super().__init__(**kwargs)
        self.poll = poll
        self.timeline_id = timeline_id
        self.post_id = post_id
        self.add_class("poll-container")
        self.styles.height = "auto"

    def compose(self):
        # The PollWidget itself acts as a vertical container for its children.
        if self.poll.get("voted") or self.poll.get("expired"):
            # Show results
            total_votes = self.poll.get("votes_count", 0)
            yield Static("Poll Results:", classes="poll-header")
            for i, option in enumerate(self.poll["options"]):
                votes = option.get("votes_count", 0)
                percentage = (votes / total_votes * 100) if total_votes > 0 else 0
                
                is_own_vote = i in self.poll.get('own_votes', [])
                
                label_prefix = "✓ " if is_own_vote else "  "
                label = f"{label_prefix}{option['title']} ({votes} votes, {percentage:.2f}%)"
                
                bar_color = "green" if is_own_vote else "blue"
                result_bar = Static(f"[{bar_color}] {'█' * int(percentage / 2)} [/]")
                
                yield Static(label)
                yield result_bar
        else:
            # Show radio buttons to vote
            yield Static("Cast your vote:", classes="poll-header")
            with RadioSet(id="poll-options"):
                for option in self.poll["options"]:
                    yield RadioButton(option["title"])

        # Add the footer with expiry and total votes
        with Horizontal(classes="poll-footer"):
            total_votes = self.poll.get("votes_count", 0)
            yield Static(f"{total_votes} votes", classes="poll-total-votes")

            expires_at = self.poll.get("expires_at")
            if expires_at:
                expiry_str = format_datetime(expires_at)
                status_text = "Expired" if self.poll.get("expired") else "Expires"
                yield Static(f"{status_text}: {expiry_str}", classes="poll-expiry")

    @on(RadioSet.Changed)
    def on_radio_set_changed(self, event: RadioSet.Changed) -> None:
        """Handle a vote."""
        if event.radio_set.pressed_index is not None:
            # Check if already voted, just in case
            if self.poll.get("voted"):
                return
            self.post_message(VoteOnPoll(self.poll["id"], event.radio_set.pressed_index, self.timeline_id, self.post_id))
            # Disable the radio set to prevent re-voting
            event.radio_set.disabled = True


class Post(Vertical):
    """A widget to display a single post."""

    def __init__(self, post, timeline_id: str, **kwargs):
        super().__init__(**kwargs)
        self.post = post
        self.timeline_id = timeline_id
        self.add_class("timeline-item")
        status_to_display = self.post.get("reblog") or self.post
        self.created_at_str = format_datetime(status_to_display["created_at"])

    def on_mount(self):
        status_to_display = self.post.get("reblog") or self.post
        if status_to_display.get("favourited"):
            self.add_class("favourited")
        if status_to_display.get("reblogged"):
            self.add_class("reblogged")

    def compose(self):
        reblog = self.post.get("reblog")
        is_reblog = reblog is not None
        status_to_display = reblog or self.post

        if is_reblog:
            booster_display_name = self.post["account"]["display_name"]
            booster_acct = self.post["account"]["acct"]
            yield Static(
                f"🚀 Boosted by {booster_display_name} (@{booster_acct})",
                classes="boost-header",
            )

        spoiler_text = status_to_display.get("spoiler_text")
        author_display_name = status_to_display["account"]["display_name"]
        author_acct = status_to_display["account"]["acct"]
        author = f"{author_display_name} (@{author_acct})"

        panel_title = author
        panel_subtitle = ""

        if spoiler_text:
            panel_title = spoiler_text
            panel_subtitle = author

        yield Static(
            Panel(
                Markdown(get_full_content_md(status_to_display)),
                title=panel_title,
                subtitle=panel_subtitle,
                box=box.ROUNDED,
                padding=(0, 1),
            )
        )
        if status_to_display.get("poll"):
            yield PollWidget(status_to_display["poll"], timeline_id=self.timeline_id, post_id=status_to_display["id"])
            
        if config.image_support and status_to_display.get("media_attachments"):
            for media in status_to_display["media_attachments"]:
                if media["type"] == "image":
                    yield ImageWidget(media["url"], config.image_renderer)

        with Horizontal(classes="post-footer"):
            yield LoadingIndicator(classes="action-spinner")
            yield Static(
                f"Boosts: {status_to_display.get('reblogs_count', 0)}", id="boost-count"
            )
            yield Static(
                f"Likes: {status_to_display.get('favourites_count', 0)}",
                id="like-count",
            )
            yield Static(self.created_at_str, classes="timestamp")

    def show_spinner(self):
        self.query_one(".action-spinner").display = True

    def hide_spinner(self):
        self.query_one(".action-spinner").display = False

    def update_from_post(self, post):
        self.post = post
        status_to_display = self.post.get("reblog") or self.post
        
        # Update classes
        self.remove_class("favourited", "reblogged")
        if status_to_display.get("favourited"):
            self.add_class("favourited")
        if status_to_display.get("reblogged"):
            self.add_class("reblogged")

        # Update stats
        self.query_one("#boost-count").update(
            f"Boosts: {status_to_display.get('reblogs_count', 0)}"
        )
        self.query_one("#like-count").update(
            f"Likes: {status_to_display.get('favourites_count', 0)}"
        )
        self.hide_spinner()

        # Re-render the poll if it exists
        for poll_widget in self.query(PollWidget):
            poll_widget.remove()
        if status_to_display.get("poll"):
            self.mount(PollWidget(status_to_display["poll"], timeline_id=self.timeline_id, post_id=status_to_display["id"]), after=self.query_one(".post-footer"))

    def on_click(self, event: events.Click) -> None:
        event.stop()
        self.post_message(SelectPost(self))

    def get_created_at(self) -> datetime | None:
        status = self.post.get("reblog") or self.post
        if status and "created_at" in status:
            ts = status["created_at"]
            if isinstance(ts, datetime):
                return ts
            return datetime.fromisoformat(ts.replace("Z", "+00:00"))
        return None


class GapIndicator(Widget):
    """A widget to indicate a gap in the timeline."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_class("gap-indicator")

    def compose(self):
        yield Static("...")


class Notification(Widget):
    """A widget to display a single notification."""

    def __init__(self, notif, **kwargs):
        super().__init__(**kwargs)
        self.notif = notif
        self.add_class("timeline-item")

        created_at = None
        if self.notif["type"] == "mention":
            created_at = self.notif["status"]["created_at"]
        else:
            created_at = self.notif["created_at"]
        self.created_at_str = format_datetime(created_at)

    def compose(self):
        notif_type = self.notif["type"]
        author = self.notif["account"]
        author_display_name = author["display_name"]
        author_acct = f"@{author['acct']}"
        author_str = f"{author_display_name} ({author_acct})"

        if notif_type == "mention":
            status = self.notif["status"]
            spoiler_text = status.get("spoiler_text")
            panel_title = f"Mention from {author_str}"
            panel_subtitle = ""
            if spoiler_text:
                panel_title = spoiler_text
                panel_subtitle = f"Mention from {author_str}"

            yield Static(
                Panel(
                    Markdown(get_full_content_md(status)),
                    title=panel_title,
                    subtitle=panel_subtitle,
                    box=box.ROUNDED,
                    padding=(0, 1),
                )
            )
            with Horizontal(classes="post-footer"):
                yield Static(
                    f"Boosts: {status.get('reblogs_count', 0)}", id="boost-count"
                )
                yield Static(
                    f"Likes: {status.get('favourites_count', 0)}", id="like-count"
                )
                yield Static(self.created_at_str, classes="timestamp")

        elif notif_type == "favourite":
            status = self.notif["status"]
            yield Static(f"❤️ {author_str} favourited your post:")
            yield Static(
                Panel(
                    Markdown(get_full_content_md(status)),
                    box=box.ROUNDED,
                    padding=(0, 1),
                )
            )
            with Horizontal(classes="post-footer"):
                yield Static(self.created_at_str, classes="timestamp")

        elif notif_type == "reblog":
            status = self.notif["status"]
            yield Static(f"🚀 {author_str} boosted your post:")
            yield Static(
                Panel(
                    Markdown(get_full_content_md(status)),
                    box=box.ROUNDED,
                    padding=(0, 1),
                )
            )
            with Horizontal(classes="post-footer"):
                yield Static(self.created_at_str, classes="timestamp")

        elif notif_type == "follow":
            yield Static(f"👋 {author_str} followed you.")
            with Horizontal(classes="post-footer"):
                yield Static(self.created_at_str, classes="timestamp")

        elif notif_type == "poll":
            status = self.notif["status"]
            poll = status.get("poll", {})
            options = poll.get("options", [])
            total_votes = poll.get("votes_count", 0)

            yield Static("📊 A poll you participated in has ended:")

            for option in options:
                title = option.get("title", "N/A")
                votes = option.get("votes_count", 0)
                percentage = (votes / total_votes * 100) if total_votes > 0 else 0
                yield Static(f"  - {title}: {votes} votes ({percentage:.2f}%)")
            with Horizontal(classes="post-footer"):
                yield Static(self.created_at_str, classes="timestamp")

        else:
            yield Static(f"Unsupported notification type: {notif_type}")

    def on_click(self, event: events.Click) -> None:
        event.stop()
        self.post_message(SelectPost(self))

    def get_created_at(self) -> datetime | None:
        ts_str = None
        if self.notif["type"] == "mention":
            ts_str = self.notif.get("status", {}).get("created_at")
        else:
            ts_str = self.notif.get("created_at")

        if ts_str:
            if isinstance(ts_str, datetime):
                return ts_str
            return datetime.fromisoformat(ts_str.replace("Z", "+00:00"))
        return None


class SearchResult(Widget, can_focus=True):
    """A base class for search results."""
    pass


class AccountResult(SearchResult):
    """A widget to display an account search result."""
    def __init__(self, account: dict, **kwargs):
        super().__init__(**kwargs)
        self.account = account
        self.add_class("search-result")

    def compose(self):
        yield Static(f"[bold]{self.account['display_name']}[/bold] @{self.account['acct']}")
        yield Static(to_markdown(self.account['note']))


class HashtagResult(SearchResult):
    """A widget to display a hashtag search result."""
    def __init__(self, hashtag: dict, **kwargs):
        super().__init__(**kwargs)
        self.hashtag = hashtag
        self.add_class("search-result")

    def compose(self):
        yield Static(f"#{self.hashtag['name']}")
        # You can add more info here, like recent usage stats if available


class StatusResult(SearchResult):
    """A widget to display a status search result."""
    def __init__(self, status: dict, **kwargs):
        super().__init__(**kwargs)
        self.status = status
        self.add_class("search-result")

    def compose(self):
        yield Post(self.status, timeline_id="search")


from textual.screen import ModalScreen
from textual.widgets import Static
from textual.containers import VerticalScroll, Container
from mastui.widgets import Post
import logging

log = logging.getLogger(__name__)

class HashtagTimeline(ModalScreen):
    """A modal screen to display a hashtag timeline."""

    BINDINGS = [
        ("escape", "dismiss", "Close Timeline"),
    ]

    def __init__(self, hashtag: str, api, **kwargs) -> None:
        super().__init__(**kwargs)
        self.hashtag = hashtag
        self.api = api

    def compose(self):
        self.title = f"#{self.hashtag}"
        with Container(id="hashtag-timeline-dialog"):
            yield VerticalScroll(
                Static(f"Loading posts for #{self.hashtag}...", classes="status-message"),
                id="hashtag-timeline-container"
            )

    def on_mount(self):
        self.run_worker(self.load_posts, thread=True)

    def load_posts(self):
        """Load the posts for the hashtag."""
        try:
            posts = self.api.timeline_hashtag(self.hashtag)
            self.app.call_from_thread(self.render_posts, posts)
        except Exception as e:
            log.error(f"Error loading hashtag timeline: {e}", exc_info=True)
            self.app.notify(f"Error loading hashtag timeline: {e}", severity="error")
            self.dismiss()

    def render_posts(self, posts):
        """Render the posts."""
        container = self.query_one("#hashtag-timeline-container")
        container.query("*").remove()

        if not posts:
            container.mount(Static(f"No posts found for #{self.hashtag}.", classes="status-message"))
            return

        for post in posts:
            container.mount(Post(post, timeline_id="hashtag"))



from textual.message import Message
from textual.widget import Widget

class PostStatusUpdate(Message):
    """A message to update a post's status."""
    def __init__(self, post_data: dict) -> None:
        self.post_data = post_data
        super().__init__()


class ActionFailed(Message):
    """A message to indicate that an action failed."""
    def __init__(self, post_id: str) -> None:
        self.post_id = post_id
        super().__init__()


class TimelineData(Message):
    """A message to send timeline data."""
    def __init__(self, timeline_id: str, posts: list) -> None:
        self.timeline_id = timeline_id
        self.posts = posts
        super().__init__()


class FocusNextTimeline(Message):
    """A message to focus the next timeline."""
    pass


class FocusPreviousTimeline(Message):
    """A message to focus the previous timeline."""
    pass


class TimelineUpdate(Message):
    """A message to update the timeline with new posts."""
    def __init__(self, posts: list, since_id: str = None, max_id: str = None) -> None:
        self.posts = posts
        self.since_id = since_id
        self.max_id = max_id
        super().__init__()


class ViewProfile(Message):
    """A message to view a user's profile."""
    def __init__(self, account_id: str) -> None:
        self.account_id = account_id
        super().__init__()


class VoteOnPoll(Message):
    """A message to vote on a poll."""
    def __init__(self, poll_id: str, choice: int, timeline_id: str, post_id: str) -> None:
        self.poll_id = poll_id
        self.choice = choice
        self.timeline_id = timeline_id
        self.post_id = post_id
        super().__init__()


class SelectPost(Message):
    """A message to select a post in a timeline."""
    def __init__(self, post_widget: Widget) -> None:
        self.post_widget = post_widget
        super().__init__()


class ResumeTimers(Message):
    """A message to resume the timeline auto-refresh timers."""
    pass



from textual.widgets import Static
from textual import events
import httpx
from io import BytesIO
from textual_image.renderable import Image, SixelImage, HalfcellImage, TGPImage
from PIL import Image as PILImage
from mastui.config import config
import hashlib
import logging

log = logging.getLogger(__name__)


class ImageWidget(Static):
    """A widget to display an image."""

    def __init__(self, url: str, renderer: str, **kwargs):
        super().__init__("Loading image...", **kwargs)
        self.url = url
        self.renderer = renderer
        self.pil_image = None

    def on_mount(self):
        self.run_worker(self.load_image, thread=True)

    def load_image(self):
        """Loads the image from the cache or URL."""
        try:
            # Create a unique filename from the URL
            filename = hashlib.sha256(self.url.encode()).hexdigest()
            cache_path = config.image_cache_dir / filename
            log.debug(f"Image cache path: {cache_path}")

            if cache_path.exists():
                log.debug(f"Loading image from cache: {self.url}")
                image_data = cache_path.read_bytes()
            else:
                log.debug(f"Image not in cache, downloading: {self.url}")
                with httpx.stream(
                    "GET", self.url, timeout=30, verify=config.ssl_verify
                ) as response:
                    response.raise_for_status()
                    image_data = response.read()
                cache_path.write_bytes(image_data)

            self.pil_image = PILImage.open(BytesIO(image_data))
            self.app.call_from_thread(self.render_image)
        except Exception as e:
            log.error(f"Error loading image: {e}", exc_info=True)
            self.app.call_from_thread(self.show_error)

    def on_resize(self, event: events.Resize) -> None:
        """Re-render the image when the widget is resized."""
        self.render_image()

    def show_error(self):
        """Displays an error message when the image fails to load."""
        self.update("[Image load failed]")

    def render_image(self):
        """Renders the image."""
        if not self.pil_image:
            return  # Image not loaded yet

        if self.pil_image.width == 0 or self.pil_image.height == 0:
            self.show_error()
            return

        renderer_map = {
            "auto": Image,
            "sixel": SixelImage,
            "ansi": HalfcellImage,
            "tgp": TGPImage,
        }
        renderer_class = renderer_map.get(self.renderer, Image)

        width = self.size.width - 4
        if width <= 0:
            self.update("...")  # Too small to render
            return

        image = renderer_class(self.pil_image, width=width, height="auto")
        self.styles.height = "auto"
        self.update(image)



import clipman
from textual.app import App, ComposeResult
from textual.containers import Grid, Vertical
from textual.widgets import Button, Label, Input, Static, TextArea, LoadingIndicator, Header
from textual.screen import ModalScreen
from urllib.parse import urlparse

from mastui.mastodon_api import login, create_app
import logging

log = logging.getLogger(__name__)


class LoginScreen(ModalScreen):
    """Screen for user to login."""

    def compose(self) -> ComposeResult:
        self.title = "Mastui Login"
        with Grid(id="dialog"):
            yield Header(show_clock=False)
            yield Label("Mastodon Instance:", id="host_label")
            yield Input(placeholder="mastodon.social", id="host")

            yield Static()  # Spacer
            yield Button("Get Auth Link", variant="primary", id="get_auth")

            yield LoadingIndicator(classes="hidden")
            yield Static(id="status", classes="hidden")

            with Vertical(id="auth_link_container", classes="hidden"):
                yield Static(
                    "1. Link copied to clipboard! Open it in your browser to authorize.",
                    id="clipboard-status",
                )
                yield TextArea(
                    "",
                    id="auth_link",
                )
                yield Static("2. Paste the authorization code here:")
                yield Input(placeholder="Authorization Code", id="auth_code")
                yield Button("Login", variant="primary", id="login")

    def on_mount(self) -> None:
        """Called when the screen is mounted."""
        self.query_one("#host").focus()

    def clean_host(self, host_input: str) -> str:
        """Cleans the host input to be a valid domain."""
        if not host_input:
            return ""
        
        # Prepend a scheme if one isn't present, for urlparse to work correctly.
        if not host_input.startswith(('http://', 'https://')):
            host_input = 'https://' + host_input
            
        parsed_url = urlparse(host_input)
        # netloc contains the domain and potentially the port
        return parsed_url.netloc

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        status = self.query_one("#status")
        spinner = self.query_one(LoadingIndicator)

        if event.button.id == "get_auth":
            host_input_widget = self.query_one("#host")
            host_value = host_input_widget.value
            if not host_value:
                status.update("[red]Please enter a Mastodon instance.[/red]")
                status.remove_class("hidden")
                return

            cleaned_host = self.clean_host(host_value)
            if not cleaned_host:
                status.update("[red]Invalid instance name.[/red]")
                status.remove_class("hidden")
                return
            
            # Update the input field with the cleaned value for user feedback
            host_input_widget.value = cleaned_host

            spinner.remove_class("hidden")
            status.add_class("hidden")
            self.run_worker(
                lambda: self.app.call_from_thread(
                    self.on_auth_link_created, create_app(cleaned_host)
                ),
                exclusive=True,
                thread=True,
            )

        elif event.button.id == "login":
            auth_code = self.query_one("#auth_code").value
            host = self.query_one("#host").value
            if not auth_code:
                status.update("[red]Please enter the authorization code.[/red]")
                status.remove_class("hidden")
                return

            spinner.remove_class("hidden")
            status.add_class("hidden")
            self.run_worker(
                lambda: self.app.call_from_thread(
                    self.on_login_complete, login(host, auth_code)
                ),
                exclusive=True,
                thread=True,
            )

    def on_auth_link_created(self, result) -> None:
        """Callback for when the auth link is created."""
        status = self.query_one("#status")
        spinner = self.query_one(LoadingIndicator)
        spinner.add_class("hidden")
        
        try:
            auth_url, error = result

            if error:
                status.update(f"[red]Error: {error}[/red]")
                status.remove_class("hidden")
                return

            try:
                clipman.init()
                clipman.set(auth_url)
                self.query_one("#clipboard-status").update(
                    "1. Link copied to clipboard! Open it in your browser to authorize."
                )
            except clipman.exceptions.ClipmanBaseException as e:
                log.warning(f"Could not copy to clipboard: {e}")
                self.query_one("#clipboard-status").update(
                    "1. Could not copy to clipboard. Please copy the link below manually:"
                )

            auth_link_input = self.query_one("#auth_link")
            auth_link_input.text = auth_url
            auth_link_input.read_only = True
            self.query_one("#auth_link_container").remove_class("hidden")
            self.query_one("#get_auth").add_class("hidden")
            self.query_one("#host_label").add_class("hidden")
            self.query_one("#host").disabled = True
            self.query_one("#auth_code").focus()
        except Exception as e:
            log.error(f"Error in on_auth_link_created: {e}", exc_info=True)
            status.update(f"[red]An unexpected error occurred. See log for details.[/red]")
            status.remove_class("hidden")

    def on_login_complete(self, result) -> None:
        """Callback for when the login is complete."""
        api, error = result
        status = self.query_one("#status")
        spinner = self.query_one(LoadingIndicator)
        spinner.add_class("hidden")

        if api:
            self.dismiss(api)
        else:
            status.update(f"[red]Login failed: {error}[/red]")
            status.remove_class("hidden")
